from sybil_engine.module.module import Module
from sybil_engine.utils.accumulator import add_accumulator, get_value

from web3_wizzard_lib.core.utils.statistic_utils import statistic_date_string, get_statistic_writer
from web3_wizzard_lib.utils.debank_utils import debank_request

TOTAL_USD = "TotalUSD"
DEBANK_ACC_NUM = "DEBANK_ACC_NUM"


class DebankChecker(Module):
    module_name = "DEBANK_CHECKER"
    module_config = None

    cumulative_chain_sums = {}

    supported_chains = [
        'ethereum',
        'arbitrum',
        'op',
        'zksync era',
        'base',
        'bnb Chain',
        'linea',
        'polygon',
        'avalanche',
        'scroll',
        'utils',
        'fantom',
        'zora',
        'arbitrum nova',
        'celo',
        'klaytn'
    ]

    def execute(self, account, statistic_write='GOOGLE'):
        data = debank_request(account.address)

        job_name = f"debank_{statistic_date_string}"
        statistics_writer = get_statistic_writer()

        add_accumulator(DEBANK_ACC_NUM, 1)

        total_usd_value = data.get('total_usd_value', 0)
        chain_list = data.get('chain_list', [])
        chain_list = [item for item in chain_list if item['name'].lower() in self.supported_chains]
        chain_list = sorted(chain_list, key=lambda x: x['community_id'])

        chain_names = [chain.get('name') for chain in chain_list]

        statistics_writer.init_if_required(
            job_name,
            ['ADS ID', 'Wallet ID', 'Total USD Value'] + chain_names
        )

        row = [account.app_id, account.address, total_usd_value]
        for chain in chain_list:
            usd_value = chain.get('usd_value', 0)
            row.append(usd_value)

            if chain['name'] in self.cumulative_chain_sums:
                self.cumulative_chain_sums[chain['name']] += usd_value
            else:
                self.cumulative_chain_sums[chain['name']] = usd_value

        statistics_writer.write_row(job_name, row)

        add_accumulator(TOTAL_USD, total_usd_value)

        if get_value(DEBANK_ACC_NUM) == get_value("Acc Amount"):
            cumulative_row = ['SUM', '', get_value(TOTAL_USD)]
            for chain_name in chain_names:
                cumulative_row.append(self.cumulative_chain_sums.get(chain_name, 0))

            statistics_writer.write_row(job_name, cumulative_row)

    def log(self):
        return "Debank Checker"



from sybil_engine.data.networks import get_chain_instance
from sybil_engine.utils.web3_utils import init_web3

from web3_wizzard_lib.core.utils.sub_module import SubModule


class NftSubmodule(SubModule):
    def execute(self, *args):
        pass

    def log(self):
        pass

    def create_web3(self, account, chain):
        chain_instance = get_chain_instance(chain)
        return init_web3(chain_instance, account.proxy)


from sybil_engine.contract.contract import Contract
from sybil_engine.data.networks import get_ids_chain
from sybil_engine.data.tokens import get_tokens_for_chain
from sybil_engine.utils.file_loader import load_abi

abi = load_abi("resources/abi/layerbank_price.json")


class LayerBankPrice(Contract):
    def __init__(self, contract_address, web3):
        self.weth_token = get_tokens_for_chain(get_ids_chain()[web3.eth.chain_id])['WETH']
        super().__init__(contract_address, web3, abi)

    def get_underlying_price(self, token):
        return self.contract.functions.getUnderlyingPrice(token).call()

    def price_of_eth(self):
        return self.contract.functions.priceOfETH().call()



from sybil_engine.contract.contract import Contract
from sybil_engine.contract.transaction_executor import evm_transaction
from sybil_engine.utils.file_loader import load_abi

abi = load_abi("resources/abi/scroll_bridge.json")


class ScrollBridgeContract(Contract):
    def __init__(self, contract_address, web3):
        super().__init__(contract_address, web3, abi)

    @evm_transaction
    def withdraw(
            self,
            account,
            amount_wei
    ):
        txn_params = self.build_generic_data(account.address, False)
        txn_params['value'] = amount_wei

        return self.contract.functions.withdrawETH(
            amount_wei,
            0
        ).build_transaction(txn_params)



from sybil_engine.contract.contract import Contract
from sybil_engine.utils.file_loader import load_abi

abi = load_abi("resources/abi/syncswap_pool.json")


class SyncSwapPoolContract(Contract):
    def __init__(self, contract_address, web3):
        super().__init__(contract_address, web3, abi)

    def get_amount_out(self, account, amount_to_swap, token_in_address):
        sender = account.address

        contract = self.web3.eth.contract(address=self.contract_address, abi=abi)

        return contract.functions.getAmountOut(token_in_address, amount_to_swap.wei, sender).call()

    def get_pool(self, from_token, to_token):
        return self.contract.functions.getPool(
            from_token,
            to_token
        ).call()

    def balanceOf(self, account):
        return self.contract.functions.balanceOf(account.address).call()


from sybil_engine.data.contracts import get_contracts_for_chain
from sybil_engine.data.networks import get_chain_instance
from sybil_engine.utils.web3_utils import init_web3

from web3_wizzard_lib.core.contract.arena_games_contract import ArenaGamesContract
from web3_wizzard_lib.core.utils.sub_module import SubModule


class ArenaGames(SubModule):
    module_name = 'ARENA_GAMES'

    def execute(self, account, chain='LINEA'):
        chain_instance = get_chain_instance(chain)
        web3 = init_web3(chain_instance, account.proxy)

        contract_address = get_contracts_for_chain(chain)['ARENA_GAMES']
        arena_games_contract = ArenaGamesContract(contract_address, web3)

        arena_games_contract.safe_mint(account)

    def log(self):
        return "ARENA GAMES NFT"



from json import dumps

import requests
from eth_account.messages import encode_defunct


def sign_in_message(account, web3):
    message_text = "Please sign this message to confirm you are the owner of this address and Sign in to TrustGo App"
    signature = sign_msg(account, message_text, web3)
    url = 'https://mp.trustalabs.ai/accounts/check_signed_message'
    data = {
        "mode": "evm", "address": account.address,
        "message": message_text,
        "signature": signature,
        "invite_from": {"from": "0", "code": 'EPR0QD9W52I8'}}
    headers = {'Authorization': f'TOKEN null', 'Accept': 'application/json'}
    data_json = dumps(data)
    r = requests.post(url, data=data_json, headers=headers)
    if r.status_code == 200:
        res = [r.json()]
        if res[0]['code'] == 0:
            token_auth = res[0]['data']['token']
            return token_auth
    return -1


def sign_msg(wallet, message_text, web3):
    text_hex = "0x" + message_text.encode('utf-8').hex()
    text_encoded = encode_defunct(hexstr=text_hex)
    signed_message = web3.eth.account.sign_message(text_encoded, private_key=wallet.key)
    signature = signed_message.signature
    return signature.hex()


import requests
from loguru import logger
from sybil_engine.data.networks import get_chain_instance
from sybil_engine.domain.balance.balance import Erc20Balance
from sybil_engine.domain.balance.tokens import Erc20Token
from sybil_engine.module.module import Module
from sybil_engine.utils.accumulator import add_accumulator_balance, add_accumulator, get_value
from sybil_engine.utils.web3_utils import init_web3

from web3_wizzard_lib.core.utils.statistic_utils import statistic_date_string, get_statistic_writer


class LineaPOHLXP(Module):
    module_name = 'LINEA_POH_LXP'
    module_config = None

    lxp_contract = "0xd83af4fbD77f3AB65C3B1Dc4B38D7e67AEcf599A"

    def execute(self, account, chain="LINEA", statistic_write="GOOGLE"):
        poh = self.linea_poh_passed(account)

        chain_instance = get_chain_instance(chain)
        web3 = init_web3(chain_instance, account.proxy)
        erc20_token = Erc20Token(chain, self.lxp_contract, web3)
        balance = erc20_token.balance(account)

        logger.info(f"{account.address} balance is {balance}")
        job_name = f"lxp_{statistic_date_string}"
        statistics_writer = get_statistic_writer()

        statistics_writer.init_if_required(
            job_name,
            ['#', 'Address', 'rLXP', 'POH', 'LXP', 'Remark']
        )

        if poh:
            rLXP = balance
            add_accumulator_balance("Total rLXP", rLXP.wei)
        else:
            rLXP = Erc20Balance(0, 'LINEA', "LXP")

        add_accumulator_balance("Total LXP", balance.wei)
        add_accumulator("Acc Num", 1)

        row = [account.app_id, account.address, float(rLXP.readable()), poh, float(balance.readable()), ""]
        statistics_writer.write_row(job_name, row)

        if get_value("Acc Num") == get_value("Acc Amount"):
            row = [
                "TOTAL",
                "",
                float(get_value("Total rLXP").readable() / 10 ** 12),
                "",
                float(get_value("Total LXP").readable() / 10 ** 12),
                ""
            ]

            statistics_writer.write_row(job_name, row)

    def linea_poh_passed(self, account):
        result = requests.get(f"https://linea-xp-poh-api.linea.build/poh/{account.address}").json()
        return result["poh"]

    def log(self):
        return "LINEA POH LXP"



import random
from hashlib import sha256

from sybil_engine.data.contracts import get_contracts_for_chain
from sybil_engine.data.networks import get_chain_instance
from sybil_engine.module.module import Order, RepeatableModule
from sybil_engine.utils.utils import interval_to_int
from sybil_engine.utils.web3_utils import init_web3

from web3_wizzard_lib.core.contract.dmail_send_mail import DmailSend


class Dmail(RepeatableModule):
    module_name = 'DMAIL'
    module_config = 'send_dmail_config'
    allowed_chains = ['ZKSYNC', 'LINEA', 'SCROLL', 'MANTA', 'BASE']
    random_order = Order.RANDOM
    repeat_conf = 'email_amount'

    @RepeatableModule.repeatable_log
    def execute(self, chain, account):
        chain_instance = get_chain_instance(chain)
        web3 = init_web3(chain_instance, account.proxy)

        dmail_contract_address = get_contracts_for_chain(chain)["DMAIL"]

        email = sha256(str(1e11 * random.random()).encode()).hexdigest()
        theme = sha256(str(1e11 * random.random()).encode()).hexdigest()

        args = [account, email, theme]

        dmail_sender = self.get_contract_class()(dmail_contract_address, web3)

        dmail_sender.send_mail(*args)

    def log(self):
        return "SEND EMAIL"

    def get_contract_class(self):
        return DmailSend

    def parse_params(self, module_params):
        self.validate_supported_chain(module_params['chain'])

        return (module_params['chain'],)

    def order(self):
        return Order.RANDOM

    def repeats(self, module_params):
        if self.repeat_conf not in module_params:
            return 1
        else:
            return range(interval_to_int(module_params[self.repeat_conf]))



import glob
import importlib
import os

import web3_wizzard_lib

from sybil_engine.module.modules import *
from loguru import logger
from sybil_engine.app import launch_with_data

from web3_wizzard_lib.core.modules.swap.swap_list import swap_facade
import importlib.util


def import_all_modules(package):
    # Get the directory path of the package
    package_dir = os.path.dirname(package.__file__)

    # Get all .py files in that directory
    python_files = glob.glob(package_dir + "/*.py")

    for file in python_files:
        # Get module name from file name by removing the extension
        module_name = os.path.basename(file)[:-3]
        # Form the absolute module path by 'package.module'
        absolute_module_name = f'{package.__name__}.{module_name}'
        # Import the module using its absolute path
        importlib.import_module(absolute_module_name)


def launch():
    import_all_modules(web3_wizzard_lib.core.modules)

    logger.info(Module.__subclasses__())

    modules_data = Modules(None, swap_facade)
    #modules_data = load_module_vars("web3-wizzard-lib.utils.modules")['modules_data']

    launch_with_data(modules_data)



from datetime import datetime

from sybil_engine.contract.contract import Contract
from sybil_engine.contract.transaction_executor import evm_transaction
from sybil_engine.data.networks import get_ids_chain
from sybil_engine.data.tokens import get_tokens_for_chain
from sybil_engine.utils.file_loader import load_abi

swap_method_id = 'c04b8d59'

length_3c = '000000000000000000000000000000000000000000000000000000000000003c'
universal_value = '000000000000000000000000000000000000000000000000000000000000002000000000000000000000000000000000000000000000000000000000000000a0'

abi = load_abi("resources/abi/maverick_multicall.json")


class MaverickMulticall(Contract):
    def __init__(self, contract_address, web3):
        self.weth_token = get_tokens_for_chain(get_ids_chain()[web3.eth.chain_id])['WETH']
        super().__init__(contract_address, web3, abi)

    @evm_transaction
    def multicall(self, account, amount_to_swap, amount_out_min_in, token_in_address, pool_address, token_out_address):
        txn_params = self.build_generic_data(account.address, False)

        if amount_to_swap.token == 'ETH':
            txn_params['value'] = amount_to_swap.wei  # amount in for ETH

        deadline_hex = decimal_to_padded_hexadecimal(int(datetime.now().timestamp() + 60 * 60 * 3), 64)
        amount_in = decimal_to_padded_hexadecimal(amount_to_swap.wei, 64)
        amount_out_min = decimal_to_padded_hexadecimal(amount_out_min_in, 64)
        recipient = pad_hex(account.address, 64) if token_out_address != self.weth_token else pad_hex('', 64)

        eth_to_usdc = '0x' + swap_method_id + universal_value + recipient + deadline_hex + amount_in + amount_out_min + length_3c + pad_hex(
            token_in_address, 40) + pad_hex(pool_address, 40) + pad_hex(token_out_address, 40) + '00000000'

        if token_out_address == self.weth_token:
            encoded_action = self.contract.encode_abi("unwrapWETH9", args=(amount_out_min_in, account.address))
        else:
            encoded_action = self.contract.encode_abi("refundETH") + pad_hex(account.address, 128)

        return self.contract.functions.multicall([eth_to_usdc, encoded_action]).build_transaction(txn_params)


def decimal_to_padded_hexadecimal(decimal_number, desired_length):
    hex_value = hex(decimal_number)[2:]

    return hex_value.zfill(desired_length)


def pad_hex(un_padded_hex, desired_length):
    return un_padded_hex[2:].zfill(desired_length)



from sybil_engine.contract.contract import Contract
from sybil_engine.contract.transaction_executor import evm_transaction
from sybil_engine.utils.file_loader import load_abi

abi = load_abi("resources/abi/imagine.json")


class ImagineContract(Contract):
    def __init__(self, contract_address, web3):
        super().__init__(contract_address, web3, abi)

    @evm_transaction
    def mint(self, account, uri):
        amount = self.web3.to_wei(0.0001, "ether")
        txn_params = self.build_generic_data(account.address, set_contract_address=False)
        txn_params['value'] = amount

        return self.contract.functions.mint(uri).build_transaction(txn_params)



import time


def benchmark(func):
    def wrap(*args, **kwargs):
        start = time.perf_counter()
        result = func(*args, **kwargs)
        end = time.perf_counter()
        print(f"{func.__name__} executed in {end - start:.4f} seconds")
        return result

    return wrap



from loguru import logger
from sybil_engine.data.networks import get_chain_instance
from sybil_engine.module.module import Module
from sybil_engine.utils.app_account_utils import AppAccount
from sybil_engine.utils.web3_utils import init_web3

from web3_wizzard_lib.core.modules.claims.layerzero_claimer import claimer_zro
from web3_wizzard_lib.core.modules.claims.rabby_claimer import claimer_rabby

native_prices = {
    'ETH': 3700,
    'USDT': 1
}


class Claimer(Module):
    module_name = 'CLAIMER'
    sleep_after_conf = True
    module_config = 'claimer_config'

    def execute(self, chain, layerzero_token, project, account: AppAccount):
        logger.info(f"CLAIM L0 tokens for {account} on {chain}")
        chain_instance = get_chain_instance(chain)
        web3 = init_web3(chain_instance, None)

        if project == 'ZRO':
            claimer_zro(account, chain_instance, layerzero_token, web3)
        else:
            claimer_rabby(account, web3)

    def log(self):
        return "CLAIMER"

    def parse_params(self, module_params):
        if 'project' not in module_params:
            module_params['project'] = 'ZRO'

        return [
            module_params['chain'],
            module_params['token'],
            module_params['project']
        ]

    def sleep_after(self):
        return self.sleep_after_conf



import time

from sybil_engine.contract.contract import Contract
from sybil_engine.contract.transaction_executor import evm_transaction
from sybil_engine.utils.file_loader import load_abi

from web3_wizzard_lib.core.modules.stargate.stargate_balance_utils import NativeBalance, Erc20Balance

abi = load_abi("resources/abi/pancake/router.json")


class PancakeRouter(Contract):
    def __init__(self, contract_address, web3):
        super().__init__(contract_address, web3, abi)

    @evm_transaction
    def swap_to_token(self, account, amount: NativeBalance, from_token: str, to_token: str, min_amount_out: int):
        txn_params = self.build_generic_data(account.address, False)

        txn_params['value'] = amount.wei

        deadline = int(time.time()) + 1000000

        transaction_data = self.contract.encode_abi(
            "exactInputSingle",
            args=[(
                from_token,
                to_token,
                500,
                account.address,
                amount.wei,
                min_amount_out,
                0
            )]
        )

        return self.contract.functions.multicall(deadline, [transaction_data]).build_transaction(txn_params)

    @evm_transaction
    def swap_to_eth(self, account, amount: Erc20Balance, from_token: str, to_token: str, min_amount_out: int):
        txn_params = self.build_generic_data(account.address, False)

        deadline = int(time.time()) + 1000000

        transaction_data = self.contract.encode_abi(
            fn_name="exactInputSingle",
            args=[(
                from_token,
                to_token,
                500,
                "0x0000000000000000000000000000000000000002",
                amount.wei,
                min_amount_out,
                0
            )]
        )

        unwrap_data = self.contract.encode_abi(
            fn_name="unwrapWETH9",
            args=[
                min_amount_out,
                account.address
            ]

        )

        contract_txn = self.contract.functions.multicall(
            deadline,
            [transaction_data, unwrap_data]
        ).build_transaction(txn_params)

        return contract_txn



import random

from loguru import logger
from sybil_engine.data.contracts import get_contracts_for_chain
from sybil_engine.data.networks import get_chain_instance
from sybil_engine.utils.web3_utils import init_web3

from web3_wizzard_lib.core.contract.zks import ZKS
from web3_wizzard_lib.core.utils.sub_module import SubModule


class ZksDomain(SubModule):
    module_name = 'TAVAERA'

    def execute(self, account, chain='ZKSYNC'):
        chain_instance = get_chain_instance(chain)
        web3 = init_web3(chain_instance, account.proxy)

        contract_address = get_contracts_for_chain(chain)['ZKS']
        zks = ZKS(contract_address, web3)

        domain_name = self.get_random_name(zks)

        zks.register(account, domain_name)

    def get_random_name(self, zks):
        domain_name = "".join(random.sample([chr(i) for i in range(97, 123)], random.randint(7, 15)))

        logger.info(f"Mint {domain_name}.zks domain")

        check_name = zks.available(domain_name)

        if check_name:
            return domain_name

        logger.info(f"{domain_name}.zks is unavailable, try another domain")

        return self.get_random_name(zks)

    def log(self):
        return "REGISTER RANDOM ZKS DOMAIN"



from sybil_engine.contract.contract import Contract
from sybil_engine.contract.transaction_executor import evm_transaction
from sybil_engine.utils.file_loader import load_abi
from web3 import Web3

abi = load_abi("resources/abi/era_ns.json")


class EraName(Contract):
    def __init__(self, contract_address, web3):
        super().__init__(contract_address, web3, abi)

    @evm_transaction
    def register(self, account, domain_name):
        sender = account.address

        txn_params = self.build_generic_data(sender, False)
        txn_params['value'] = Web3.to_wei(0.003, "ether")

        return self.contract.functions.Register(domain_name).build_transaction(txn_params)

    def check_name(self, domain_name):
        return self.contract.functions._checkName(domain_name).call()
